### #define SYS_clone 220

* 功能：创建一个子进程；
* 输入：
  - flags: 创建的标志，如SIGCHLD；
  - stack: 指定新进程的栈，可为0；
  - ptid: 父线程ID；
  - tls: TLS线程本地存储描述符；
  - ctid: 子线程ID；
* 返回值：成功则返回子进程的线程ID，失败返回-1；

```
pid_t ret = syscall(SYS_clone, flags, stack, ptid, tls, ctid)
```

对于这个syscall我们添加了一个clone函数,保存在proc.c中,这个clone函数基本上和fork函数只有一处区别,在保存栈帧的时候把栈帧中的sp变量复制为传入的stack即可.

具体的函数请参见`sys_clone()`.大概的思路是获取元素,然后直接调用clone()函数.

### #define SYS_execve 221

* 功能：执行一个指定的程序；
* 输入：
  - path: 待执行程序路径名称，
  - argv: 程序的参数， 
  - envp: 环境变量的数组指针
* 返回值：成功不返回，失败返回-1；

```
const char *path, char *const argv[], char *const envp[];
int ret = syscall(SYS_execve, path, argv, envp);
```

具体的函数参见`sys_execve()`.大概的思路就是先获取文件的路径,然后进行一系列ELF的处理,然后调用内核提供的exec函数即可.

### #define SYS_wait4 260

* 功能：等待进程改变状态;
* 输入：
  - pid: 指定进程ID，可为-1等待任何子进程；
  - status: 接收状态的指针；
  - options: 选项：WNOHANG，WUNTRACED，WCONTINUED；
* 返回值：成功则返回进程ID；如果指定了WNOHANG，且进程还未改变状态，直接返回0；失败则返回-1；

```
pid_t pid, int *status, int options;
pid_t ret = syscall(SYS_wait4, pid, status, options);
```

对于这个syscall我们添加了一个wait4pid函数,保存在proc.c中,这个waitpid函数基本上和wait函数只有一处区别,在判断等待与被等待的关系的if判定中添加对于pid的限制即可.

具体的函数参见`sys_wait4()`.大概的思路是首先先获取相关的参数,如果等待的进程pid为-1,就退化为wait的调用,如果不是,就调用wait4pid,然后返回值给p,这里为了保险是按字节进行返回的.

### #define SYS_exit 93

* 功能：触发进程终止，无返回值；
* 输入：终止状态值；
* 返回值：无返回值；

```
int ec;
syscall(SYS_exit, ec);
```

具体的函数参见`sys_exit()`.大概的思路是获取相关的参数后直接调用exit()函数.

### #define SYS_getppid 173

* 功能：获取父进程ID；
* 输入：系统调用ID；
* 返回值：成功返回父进程ID；

```
pid_t ret = syscall(SYS_getppid);
```

具体的函数参见`sys_getppid()`.直接获取进程结构体的parent字段的pid值即可.

### #define SYS_getpid 172

* 功能：获取进程ID；
* 输入：系统调用ID；
* 返回值：成功返回进程ID；

```
pid_t ret = syscall(SYS_getpid);
```

具体函数参见`sys_getpid()`.直接获取进程的pid值即可.

## 