// Timer Interrupt handler


#include "include/types.h"
#include "include/param.h"
#include "include/riscv.h"
#include "include/string.h"
#include "include/spinlock.h"
#include "include/vm.h"
#include "include/printf.h"
#include "include/proc.h"
#include "include/elf.h"

char shstrtab[300];
struct sechdr dynstr;
char strbuf[300];

void
secdisplay(struct sechdr* sec)
{
  char *name = shstrtab+sec->name;
  printf("-------section %s-------\n",name);
  printf("type:%p\n",sec->type);
  printf("flags:%p\n",sec->flags);
  printf("addr:%p\n",sec->addr);
  printf("offset:%p\n",sec->offset);
  printf("size:%p\n",sec->size);
  printf("link:%p\n",sec->link);
  printf("info:%p\n",sec->info);
  printf("addralign:%p\n",sec->addralign);
  printf("entsize:%p\n",sec->entsize);
  printf("-------section %s-------\n",name);
}

void symentdisplay(struct syment* ent){
  printf("-------sym -------\n");
  printf("name:%p\n",ent->name);
  printf("info:%p\n",ent->info);
  printf("other:%p\n",ent->other);
  printf("shndx:%p\n",ent->shndx);
  printf("value:%p\n",ent->value);
  printf("size:%p\n",ent->size);
  printf("-------sym -------\n");
}

uint64
loadstrtab(struct dirent* ep,struct sechdr* strsec,char* buf)
{
  if(eread(ep, 0, (uint64)buf, strsec->offset, strsec->size) != strsec->size)
    return -1;
  return 0;
}

uint64
visitstrtab(struct dirent* ep,struct sechdr* strsec,uint64 off,char* buf){
  if(eread(ep, 0, (uint64)buf, strsec->offset+off, 50) != 50)
    return -1;
  return 0;
}

uint64
loadent(struct dirent* ep,struct sechdr* sec,int id,void* entry)
{
  uint64 off = id*sec->entsize;
  if(eread(ep, 0, (uint64)entry, sec->offset+off, sec->entsize) != sec->entsize)
    return -1;
  return off;
}


void
relaentdisplay(struct dirent* ep,struct sechdr* dynsym,struct relaent* ent)
{
  struct syment sym;
  uint64 id = INFO2SYMIDX(ent->info);
  if(id){
    if(loadent(ep,dynsym,id,&sym)<0){
      printf("load dynsym failed\n");
      return;
    }
    if(visitstrtab(ep,&dynstr,sym.name,strbuf)<0){
      printf("sym name not found\n");
      return;
    }
    //symentdisplay(&sym);
  }
  printf("%p\t",ent->offset);
  printf("%p\t",ent->info);
  if(id)printf("%s + ",strbuf);
  if(ent->addend)printf("%p\n",ent->addend);
  else printf("0\n");
}

uint64
visitrela(pagetable_t pagetable,struct dirent* ep,struct sechdr* dynsym,struct sechdr* sec)
{
  struct relaent ent;
  //printf("section %s have %d entries\n",shstrtab+sec->name,sec->size/sec->entsize);
  //printf("offset\t\t\tinfo\t\t\taddend\n");
  for(int i = 0;;i++){
    uint64 off = loadent(ep,sec,i,&ent);
    if(off==-1){
      return -1;
    }
    //relaentdisplay(ep,dynsym,&ent);
    struct syment sym;
    uint64 id = INFO2SYMIDX(ent.info);
    if(id){
      if(loadent(ep,dynsym,id,&sym)<0){
        printf("load dynsym failed\n");
        return -1;
      }
      //symentdisplay(&sym);
    }
    uint64 originva = ent.offset;
    uint64 originvalue = id?sym.value:0+ent.addend;
    uint64* pa = (void*)kwalkaddr1(pagetable,linkermap(originva));
    if(!pa){
      printf("rela addr not mapped\n");
    }
    *pa = linkermap(originvalue);
    //printf("addr:%p value:%p\n",originva,originvalue);
    if(off+sec->entsize>=sec->size){
      break;
    }
  }
  //printf("\n");
  return 0;
}

uint64
loadsec(pagetable_t pagetable,struct dirent* ep,struct elfhdr* elf)
{
  //printf("shstrndx:%d\n",elf->shstrndx);
  struct sechdr shstrsec;
  struct sechdr rela_dyn_sec;
  struct sechdr rela_plt_sec;
  struct sechdr dynsym;
  //struct sechdr strtab;
  if(eread(ep, 0, (uint64)&shstrsec, elf->shoff+elf->shstrndx*elf->shentsize, sizeof(shstrsec)) != sizeof(shstrsec))
    return -1;
  if(loadstrtab(ep,&shstrsec,shstrtab)<0){
    return -1;
  }
  for(int i = 0;i<elf->shnum;i++){
    uint64 off = elf->shoff+i*elf->shentsize;
    struct sechdr sec;
    char* secname;
    if(eread(ep, 0, (uint64)&sec, off, sizeof(sec)) != sizeof(sec))
      return -1;
    secname = shstrtab+sec.name;
    //secdisplay(&sec);
    if(strncmp(".rela.dyn",secname,300)==0){
      memmove(&rela_dyn_sec,&sec,sizeof(sec));
    }else if(strncmp(".rela.plt",secname,300)==0){
      memmove(&rela_plt_sec,&sec,sizeof(sec));
    }else if(strncmp(".dynsym",secname,300)==0){
      memmove(&dynsym,&sec,sizeof(sec));
    }else if(strncmp(".dynstr",secname,300)==0){
      memmove(&dynstr,&sec,sizeof(sec));
    }
  }
  /*
  secdisplay(&rela_dyn_sec);
  secdisplay(&rela_plt_sec);
  secdisplay(&dynsym);
  secdisplay(&dynstr);
  */
  if(visitrela(pagetable,ep,&dynsym,&rela_dyn_sec)==-1){
    return -1;
  }
  if(visitrela(pagetable,ep,&dynsym,&rela_plt_sec)==-1){
    return -1;
  }
  return 0;
}

